# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Tests for the HPC executor plugin."""

import asyncio
import os
import subprocess
from copy import deepcopy
from datetime import timedelta
from functools import partial
from pathlib import Path
from unittest import mock

import asyncssh
import cloudpickle as pickle
import pytest
from covalent._results_manager.result import Result
from covalent._shared_files.config import get_config, set_config
from covalent._workflow.transport import TransportableObject
from covalent.executor.base import wrapper_fn

from covalent_hpc_plugin import HPCExecutor


@pytest.fixture
def proc_mock():
    return mock.Mock()


@pytest.fixture
def conn_mock():
    return mock.Mock()


def mock_key_read(*args, **kwargs):
    """Mock for asyncssh.read_private_key() and asyncssh.read_certificate()"""
    return True


def mock_basic_async(*args, **kwargs):
    """A basic async mock"""
    future = asyncio.Future()
    future.set_result(True)
    return future


def mock_sleep(*args, **kwargs):
    future = asyncio.Future()
    future.set_exception(ValueError("No sleep!"))
    return future


def mock_fetch_result(
    result: any = "result", stdout: str = "", stderr: str = "", exception: Exception = None
):
    """Mock for HPCExecutor._fetch_result()"""

    def mock_func(*args, **kwargs):
        future = asyncio.Future()
        future.set_result((result, stdout, stderr, exception))
        return future

    return mock_func


def mock_asyncssh_run(returncode: int = 0, stdout: str = "", stderr: str = ""):
    """Mock for asyncssh.SSHClientConnection.run()"""

    def mock_func(*args, **kwargs):
        class MockStdout:
            def __init__(self):
                self.returncode = returncode
                self.stdout = stdout
                self.stderr = stderr

        future = asyncio.Future()
        future.set_result(MockStdout())
        return future

    return mock_func


def test_init_defaults(tmpdir):
    """Test that initialization properly sets member variables."""
    tmpdir.chdir()

    # Test with defaults
    address = "host"
    executor = HPCExecutor(address=address)
    assert executor.address == address
    assert executor.username == ""
    assert executor.ssh_key_file == "~/.ssh/id_rsa"
    assert executor.cert_file is None
    assert executor.instance == "slurm"
    assert executor.inherit_environment == True
    assert executor.environment == {}
    assert executor.resource_spec_kwargs == {
        "node_count": 1,
        "exclusive_node_use": False,
        "process_count": 1,
        "processes_per_node": 1,
        "cpu_cores_per_process": 1,
        "gpu_cores_per_process": 0,
    }
    assert executor.job_attributes_kwargs == {"duration": timedelta(minutes=10)}
    assert executor.launcher == "single"
    assert executor.remote_workdir == "~/covalent-workdir"
    assert executor.create_unique_workdir == False
    assert executor.cache_dir == str(
        Path(get_config("dispatcher.cache_dir")).expanduser().resolve()
    )
    assert executor.poll_freq == 60
    assert executor.cleanup is True


def test_init_nondefaults(tmpdir):
    """Test that initialization properly sets member variables."""

    tmpdir.chdir()

    # Test with non-defaults
    address = "host"
    username = "username"
    executor = HPCExecutor(
        address=address,
        username=username,
        ssh_key_file="ssh_key_file",
        cert_file="cert_file",
        instance="flux",
        inherit_environment=False,
        environment={"hello": "world"},
        resource_spec_kwargs={"node_count": 2},
        job_attributes_kwargs={"duration": 20},
        launcher="multiple",
        remote_python_exe="python2",
        remote_conda_env="myenv",
        remote_workdir="~/my-remote-dir",
        create_unique_workdir=True,
        cache_dir=tmpdir / "my-cache-dir",
        poll_freq=90,
        cleanup=False,
    )
    assert executor.username == username
    assert executor.address == address
    assert executor.ssh_key_file == "ssh_key_file"
    assert executor.cert_file == "cert_file"
    assert executor.instance == "flux"
    assert executor.inherit_environment == False
    assert executor.environment == {"hello": "world"}
    assert executor.resource_spec_kwargs == {"node_count": 2}
    assert executor.job_attributes_kwargs == {"duration": 20}
    assert executor.launcher == "multiple"
    assert executor.remote_python_exe == "python2"
    assert executor.remote_conda_env == "myenv"
    assert executor.remote_workdir == "~/my-remote-dir"
    assert executor.create_unique_workdir == True
    assert executor.cache_dir == tmpdir / "my-cache-dir"
    assert executor.poll_freq == 90
    assert executor.cleanup is False
    assert os.path.exists(tmpdir / "my-cache-dir")


def test_init_defaults(tmpdir):
    """Test that initialization properly sets member variables."""
    tmpdir.chdir()

    # Test poll freq is auto-raised
    executor = HPCExecutor(address="host", poll_freq=10)
    assert executor.poll_freq == 30


def test_removed_inits(tmpdir):
    """Test for removed inits"""

    tmpdir.chdir()

    start_config = deepcopy(get_config())
    for key in ["cert_file", "remote_conda_env"]:
        config = get_config()
        config["executors"]["hpc"].pop(key, None)
        set_config(config)
        executor = HPCExecutor(address="host")
        assert not executor.__dict__[key]
        set_config(start_config)


def test_format_pickle_script(tmpdir):
    """Test that the python script (in string form) for the pickle function is as-expected"""

    tmpdir.chdir()

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        cert_file="cert_file",
        remote_workdir="/federation/test_user/.cache/covalent",
        cache_dir="~/.cache/covalent",
    )

    dispatch_id = "148dedae-1b58-3870-z08d-db89bceec915"
    task_id = 2
    func_filename = f"func-{dispatch_id}-{task_id}.pkl"
    result_filename = f"result-{dispatch_id}-{task_id}.pkl"
    executor._remote_func_filepath = func_filename
    executor._remote_result_filepath = result_filename

    py_script_str = executor._format_pickle_script()
    assert func_filename in py_script_str
    assert result_filename in py_script_str


def test_pickle_script(tmpdir):
    """Test Python pickle script works appropriately"""

    tmpdir.chdir()

    def test_func(a, b="default"):
        return f"{a} {b}"

    executor = HPCExecutor(username="test_user", address="test_address")
    dispatch_id = "148dedae-1b58-3870-z08d-db89bceec915"
    task_id = 2
    func_filename = tmpdir / f"func-{dispatch_id}-{task_id}.pkl"
    result_filename = tmpdir / f"result-{dispatch_id}-{task_id}.pkl"
    executor._remote_func_filepath = func_filename
    pickle.dump([test_func, {"hello"}, {"b": "world"}], open(func_filename, "wb"))
    executor._remote_result_filepath = result_filename

    py_script_str = executor._format_pickle_script()
    with open("test.py", "w") as w:
        w.write(py_script_str)
    p = subprocess.run(
        "python test.py", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    assert p.returncode == 0
    assert p.stderr == b""
    assert os.path.exists(result_filename)
    pickle_load = pickle.load(open(result_filename, "rb"))
    assert pickle_load[0] == "hello world"
    assert pickle_load[1] is None


def test_format_submit_script(tmpdir):
    """Test that the shell script (in string form) which is to be submitted on
    the remote server is created with no errors."""

    tmpdir.chdir()

    remote_workdir = "/federation/test_user/.cache/covalent"
    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="~/.ssh/id_rsa",
        remote_workdir=remote_workdir,
        remote_python_exe="python3",
        environment={"hello": "world"},
        launcher="srun",
        resource_spec_kwargs={"node_count": 10},
        job_attributes_kwargs={"duration": timedelta(minutes=20)},
        remote_conda_env="myenv",
    )

    dispatch_id = "259efebf-2c69-4981-a19e-ec90cdffd026"
    task_id = 3
    executor._name = f"{dispatch_id}-{task_id}"
    executor._remote_pickle_script_filepath = f"script-{dispatch_id}-{task_id}.py"
    executor._job_remote_workdir = tmpdir
    executor._remote_stdout_filepath = f"stdout-{dispatch_id}-{task_id}.log"
    executor._remote_stderr_filepath = f"stderr-{dispatch_id}-{task_id}.log"
    executor._remote_pre_launch_filepath = f"pre-launch-{dispatch_id}-{task_id}.sh"

    submit_script_str = executor._format_job_script()

    assert "JobSpec" in submit_script_str
    assert f'name="{executor._name}"' in submit_script_str
    assert f'executable="python3"' in submit_script_str
    assert (
        "directory=" in submit_script_str
        and str(executor._job_remote_workdir) in submit_script_str
    )
    assert "environment={'hello': 'world'}" in submit_script_str
    assert (
        "stdout_path=" in submit_script_str
        and executor._remote_stdout_filepath in submit_script_str
    )
    assert (
        "stderr_path=" in submit_script_str
        and executor._remote_stderr_filepath in submit_script_str
    )
    assert 'launcher="srun"' in submit_script_str
    assert "resources=ResourceSpecV1(**{'node_count': 10})" in submit_script_str
    assert (
        "attributes=JobAttributes(**{'duration': datetime.timedelta(seconds=1200)})"
        in submit_script_str
    )
    assert (
        "pre_launch=" in submit_script_str
        and executor._remote_pre_launch_filepath in submit_script_str
    )

def test_datetime(tmpdir):
    tmpdir.chdir()
    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="~/.ssh/id_rsa",
        job_attributes_kwargs={"duration": 20},
    )
    dispatch_id = "259efebf-2c69-4981-a19e-ec90cdffd026"
    task_id = 3
    executor._name = f"{dispatch_id}-{task_id}"
    executor._remote_pickle_script_filepath = f"script-{dispatch_id}-{task_id}.py"
    executor._job_remote_workdir = tmpdir
    executor._remote_stdout_filepath = f"stdout-{dispatch_id}-{task_id}.log"
    executor._remote_stderr_filepath = f"stderr-{dispatch_id}-{task_id}.log"
    executor._remote_pre_launch_filepath = f"pre-launch-{dispatch_id}-{task_id}.sh"

    submit_script_str = executor._format_job_script()
    assert (
        "attributes=JobAttributes(**{'duration': datetime.timedelta(seconds=1200)})"
        in submit_script_str
    )
    
def test_format_submit_script_minimal(tmpdir):
    """Test that the shell script (in string form) which is to be submitted on
    the remote server is created with no errors."""

    tmpdir.chdir()
    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        resource_spec_kwargs={},
        job_attributes_kwargs={},
    )

    dispatch_id = "259efebf-2c69-4981-a19e-ec90cdffd026"
    task_id = 3
    executor._name = f"{dispatch_id}-{task_id}"
    executor._remote_pickle_script_filepath = f"script-{dispatch_id}-{task_id}.py"
    executor._job_remote_workdir = tmpdir
    executor._remote_stdout_filepath = f"stdout-{dispatch_id}-{task_id}.log"
    executor._remote_stderr_filepath = f"stderr-{dispatch_id}-{task_id}.log"
    executor._remote_pre_launch_filepath = f"pre-launch-{dispatch_id}-{task_id}.sh"

    submit_script_str = executor._format_job_script()
    assert "resources" not in submit_script_str
    assert "attributes" not in submit_script_str
    assert "pre_launch" in submit_script_str


def test_submit_script(tmpdir):
    """Test that the submit script actually works as intended"""
    tmpdir.chdir()

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="local",
    )
    dispatch_id = "259efebf-2c69-4981-a19e-ec90cdffd026"
    task_id = 3
    executor._name = f"{dispatch_id}-{task_id}"
    executor._remote_pickle_script_filepath = f""
    executor._job_remote_workdir = tmpdir
    executor._remote_stdout_filepath = f"stdout-{dispatch_id}-{task_id}.log"
    executor._remote_stderr_filepath = f"stderr-{dispatch_id}-{task_id}.log"
    executor._remote_pre_launch_filepath = f"pre-launch-{dispatch_id}-{task_id}.sh"

    submit_str = executor._format_job_script()
    assert f'JobExecutor.get_instance("local")' in submit_str

    with open("test_submit.py", "w") as w:
        w.write(submit_str)
    p = subprocess.run(
        "python test_submit.py", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    assert p.returncode == 0
    assert p.stderr == b""
    assert p.stdout != b""
    assert int(p.stdout) > 0


def test_format_query_script():
    """Test that the Python script to perform job queries is formatted correctly"""
    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="flux",
    )

    executor._jobid = "123456"
    query_str = executor._format_query_status_script()
    assert f'JobExecutor.get_instance("flux")' in query_str
    assert f'job_executor.attach(job, f"{executor._jobid}")' in query_str


def test_query_script(tmpdir):
    """Test that the Python script to perform job queries works as expected"""
    tmpdir.chdir()

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="local",
    )
    dispatch_id = "259efebf-2c69-4981-a19e-ec90cdffd026"
    task_id = 3
    executor._name = f"{dispatch_id}-{task_id}"
    executor._remote_pickle_script_filepath = f""
    executor._job_remote_workdir = tmpdir
    executor._remote_stdout_filepath = f"stdout-{dispatch_id}-{task_id}.log"
    executor._remote_stderr_filepath = f"stderr-{dispatch_id}-{task_id}.log"
    executor._remote_pre_launch_filepath = f"pre-launch-{dispatch_id}-{task_id}.sh"

    submit_str = executor._format_job_script()
    assert f'JobExecutor.get_instance("local")' in submit_str

    executor._jobid = "{native_id}"
    query_str = executor._format_query_status_script()
    with open("test_query.py", "w") as w:
        w.write(submit_str + query_str)
    p = subprocess.run(
        "python test_query.py", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    assert p.returncode == 0
    assert p.stderr == b""
    assert "ACTIVE" in p.stdout.decode()


def test_format_pre_launch_script(tmpdir):
    """Test that the prelaunch script is formatted appropriately"""
    tmpdir.chdir()

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="flux",
    )

    executor._remote_python_version = "3.8.5"

    pre_launch_str = executor._format_pre_launch_script()
    assert pre_launch_str.split("\n")[0] == "#!/bin/bash"
    assert "3.8.5" in pre_launch_str
    assert "conda" not in pre_launch_str

    executor = HPCExecutor(
        username="test_user", address="test_address", instance="flux", remote_conda_env="myenv"
    )
    executor._remote_python_version = "3.8.5"

    pre_launch_str = executor._format_pre_launch_script()
    assert "3.8.5" in pre_launch_str
    assert "conda activate myenv" in pre_launch_str

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="flux",
        remote_conda_env="myenv",
        pre_launch_cmds=["echo hello", "echo world"],
    )
    executor._remote_python_version = "3.8.5"

    pre_launch_str = executor._format_pre_launch_script()
    assert pre_launch_str.split("\n")[0] == "#!/bin/bash"
    assert "3.8.5" in pre_launch_str
    assert "conda activate myenv" in pre_launch_str
    assert "echo hello\necho world\n" in pre_launch_str

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="flux",
        remote_conda_env="myenv",
        pre_launch_cmds=["echo hello", "echo world"],
        shebang=None,
    )
    executor._remote_python_version = "3.8.5"

    pre_launch_str = executor._format_pre_launch_script()
    assert "#!/bin/bash" not in pre_launch_str
    assert "3.8.5" in pre_launch_str
    assert "conda activate myenv" in pre_launch_str
    assert "echo hello\necho world\n" in pre_launch_str


def test_format_post_launch_script(tmpdir):
    """Test that the postlaunch script is formatted appropriately"""
    tmpdir.chdir()
    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="flux",
        post_launch_cmds=["echo hello", "echo world"],
    )
    post_launch_str = executor._format_post_launch_script()
    assert post_launch_str.split("\n")[0] == "#!/bin/bash"
    assert "echo hello\necho world\n" in post_launch_str
    assert "conda" not in post_launch_str

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="flux",
        post_launch_cmds=["echo hello", "echo world"],
        remote_conda_env="myenv",
    )
    post_launch_str = executor._format_post_launch_script()
    assert post_launch_str.split("\n")[0] == "#!/bin/bash"
    assert "echo hello\necho world\n" in post_launch_str
    assert "conda activate myenv" in post_launch_str

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        instance="flux",
        post_launch_cmds=["echo hello", "echo world"],
        remote_conda_env="myenv",
        shebang=None,
    )
    post_launch_str = executor._format_post_launch_script()
    assert "#!/bin/bash" not in post_launch_str
    assert "echo hello\necho world\n" in post_launch_str
    assert "conda activate myenv" in post_launch_str


@pytest.mark.asyncio
async def test_client_connect_failure(tmpdir, monkeypatch):
    """Test for _client_connect without mocking .connect()"""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.read_private_key", mock_key_read)
    monkeypatch.setattr("asyncssh.read_certificate", mock_key_read)

    with pytest.raises(RuntimeError):
        await HPCExecutor(
            address="test_address", username="test_use", ssh_key_file="ssh_key_file"
        )._client_connect()


@pytest.mark.asyncio
async def test_client_connect(tmpdir, monkeypatch):
    """Test for _client_connect with mocking .connect()"""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.read_private_key", mock_key_read)
    monkeypatch.setattr("asyncssh.read_certificate", mock_key_read)
    monkeypatch.setattr("asyncssh.connect", mock_basic_async)

    assert (
        await HPCExecutor(
            address="test_address", username="test_use", ssh_key_file="ssh_key_file"
        )._client_connect()
        == True
    )

    assert (
        await HPCExecutor(
            address="test_address",
            username="test_use",
            ssh_key_file="ssh_key_file",
            cert_file="cert_file",
        )._client_connect()
        == True
    )

    assert (
        await HPCExecutor(
            address="test_address", username="test_use", ssh_key_file=None, cert_file=None
        )._client_connect()
        == True
    )

    with pytest.raises(ValueError, match="address is a required parameter"):
        await HPCExecutor(ssh_key_file="ssh_key_file")._client_connect()

    with pytest.raises(ValueError, match="ssh_key_file is required if cert_file is provided."):
        await HPCExecutor(
            address="test_address", ssh_key_file=None, cert_file="cert_file"
        )._client_connect()


@pytest.mark.asyncio
async def test_get_status_success(tmpdir, monkeypatch):
    """Test the get_status method."""
    tmpdir.chdir()
    monkeypatch.setattr(
        "asyncssh.SSHClientConnection.run", mock_asyncssh_run(0, "Fake Status", "")
    )

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/federation/test_user/.cache/covalent",
    )

    # Test when jobid is not set
    status = await executor.get_status(asyncssh.SSHClientConnection)
    assert status == Result.NEW_OBJ

    # Test when jobid is set
    executor._jobid = "123456"
    executor._remote_query_script_filepath = "mock.py"
    status = await executor.get_status(asyncssh.SSHClientConnection)
    assert status == "Fake Status"


@pytest.mark.asyncio
async def test_get_status_failures(tmpdir, monkeypatch):
    """Test the get_status method for failures."""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.SSHClientConnection.run", mock_asyncssh_run(1, "", "AN ERROR"))

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/federation/test_user/.cache/covalent",
    )
    executor._jobid = "123456"
    executor._remote_query_script_filepath = "mock.py"

    with pytest.raises(RuntimeError, match="Getting job status failed: AN ERROR"):
        await executor.get_status(asyncssh.SSHClientConnection)


@pytest.mark.asyncio
async def test_poll_scheduler_completed(tmpdir, monkeypatch):
    """Test that polling the status works."""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.SSHClientConnection.run", mock_asyncssh_run(0, "COMPLETED", ""))

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/federation/test_user/.cache/covalent",
    )

    # Check completed status does not give any errors
    executor._jobid = "123456"
    executor._remote_query_script_filepath = "mock.py"
    assert await executor._poll_scheduler(asyncssh.SSHClientConnection) is None


@pytest.mark.asyncio
async def test_poll_scheduler_canceled(tmpdir, monkeypatch):
    """Test that polling the status works."""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.SSHClientConnection.run", mock_asyncssh_run(0, "CANCELED", ""))

    # Check canceled status reported
    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/federation/test_user/.cache/covalent",
    )
    executor._jobid = "12345"
    executor._remote_query_script_filepath = "mock.py"

    with pytest.raises(RuntimeError, match="Status for job with native ID 12345: CANCELED."):
        await executor._poll_scheduler(asyncssh.SSHClientConnection)


@pytest.mark.asyncio
async def test_poll_scheduler_loop(tmpdir, monkeypatch):
    """Test that queued state results in continued polling."""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.SSHClientConnection.run", mock_asyncssh_run(0, "QUEUED", ""))
    monkeypatch.setattr("asyncio.sleep", mock_sleep)

    # Check queued status works
    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/federation/test_user/.cache/covalent",
        poll_freq=1,
    )
    executor._jobid = "12345"
    executor._remote_query_script_filepath = "mock.py"

    with pytest.raises(ValueError, match="No sleep!"):
        await executor._poll_scheduler(asyncssh.SSHClientConnection)

    monkeypatch.setattr("asyncssh.SSHClientConnection.run", mock_asyncssh_run(0, "ACTIVE", ""))
    with pytest.raises(ValueError, match="No sleep!"):
        await executor._poll_scheduler(asyncssh.SSHClientConnection)


@pytest.mark.asyncio
async def test_fetch_result(tmpdir, monkeypatch):
    """Test querying results works as expected."""
    tmpdir.chdir()
    monkeypatch.setattr(
        "asyncssh.SSHClientConnection.run", mock_asyncssh_run(1, "stdout", "stderr")
    )

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/federation/test_user/.cache/covalent",
    )

    executor._remote_result_filepath = Path("/path/to/file")

    with pytest.raises(FileNotFoundError, match="stderr"):
        await executor._fetch_result(asyncssh.SSHClientConnection)


@pytest.mark.asyncio
async def test_fetch_result_v2(monkeypatch, tmpdir):
    """Test querying results works as expected."""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.SSHClientConnection.run", mock_asyncssh_run(0, "", ""))
    monkeypatch.setattr("asyncssh.scp", mock_basic_async)

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/federation/test_user/.cache/covalent",
    )
    executor._remote_result_filepath = Path("/path/to/test.txt")
    executor._task_results_dir = tmpdir
    executor._remote_stdout_filepath = Path(os.path.join(tmpdir, "stdout.txt"))
    executor._remote_stderr_filepath = Path(os.path.join(tmpdir, "stderr.txt"))
    results_file = Path(os.path.join(tmpdir, "test.txt"))
    pickle.dump(["result", "exception"], open("test.pkl", "wb"))

    with open(results_file, "wb") as f:
        with open("test.pkl", "rb") as f2:
            f.write(f2.read())
    with open(executor._remote_stdout_filepath, "w") as f:
        f.write("stdout")
    with open(executor._remote_stderr_filepath, "w") as f:
        f.write("stderr")

    outputs = await executor._fetch_result(conn=asyncssh.SSHClientConnection)
    assert outputs[0] == "result"
    assert outputs[1] == "stdout"
    assert outputs[2] == "stderr"
    assert outputs[3] == "exception"


@pytest.mark.asyncio
async def test_run(tmpdir, monkeypatch, proc_mock, conn_mock):
    """Test calling run works as expected."""
    tmpdir.chdir()
    monkeypatch.setattr("asyncssh.scp", mock_basic_async)
    monkeypatch.setattr("covalent_hpc_plugin.HPCExecutor._fetch_result", mock_fetch_result())

    executor = HPCExecutor(
        username="test_user",
        address="test_address",
        ssh_key_file="ssh_key_file",
        remote_workdir="/scratch/user/experiment1",
        create_unique_workdir=True,
        remote_conda_env="my-conda-env",
        pre_launch_cmds=["echo hello", "echo world"],
        post_launch_cmds=["echo goodbye", "echo world"],
    )

    # dummy objects
    def f(x, y):
        return x + y

    dummy_function = partial(wrapper_fn, TransportableObject(f), call_before=[], call_after=[])

    dummy_metadata = {
        "dispatch_id": "259efebf-2c69-4981-a19e-ec90cdffd026",
        "node_id": 1,
        "results_dir": "results/directory/on/remote",
    }

    dummy_args = (
        dummy_function,
        [TransportableObject(2)],
        {"y": TransportableObject(3)},
        dummy_metadata,
    )

    # mock behavior
    conn_mock.run = mock.AsyncMock(return_value=proc_mock)
    conn_mock.wait_closed = mock.AsyncMock(return_value=None)

    async def __client_connect_succeed(*_):
        return conn_mock

    # patches
    patch_ccs = mock.patch.object(HPCExecutor, "_client_connect", new=__client_connect_succeed)

    # check run/teardown method works as expected
    proc_mock.stdout = "COMPLETED"
    proc_mock.stderr = ""
    proc_mock.returncode = 0
    with patch_ccs:
        await executor.run(*dummy_args)
        await executor.teardown(dummy_metadata)

    # Test errors
    proc_mock.stderr = "FAILED"
    with patch_ccs:
        with pytest.raises(RuntimeError, match="Making remote directory failed: FAILED"):
            await executor.run(*dummy_args)

    msg = "Failed to create directory"
    proc_mock.stderr = msg
    with patch_ccs:
        with pytest.raises(RuntimeError, match=msg):
            await executor.run(*dummy_args)

    # Test job fetching error
    proc_mock.stdout = "COMPLETED"
    proc_mock.stderr = ""
    proc_mock.returncode = 0
    monkeypatch.setattr(
        "covalent_hpc_plugin.HPCExecutor._fetch_result",
        mock_fetch_result("result", "", "error", RuntimeError),
    )
    with pytest.raises(RuntimeError, match="Fetching job result failed: error"):
        with patch_ccs:
            await executor.run(*dummy_args)
    monkeypatch.setattr("covalent_hpc_plugin.HPCExecutor._fetch_result", mock_fetch_result())



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import site
import sys

from setuptools import find_packages, setup

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

plugins_list = ["hpc = covalent_hpc_plugin.hpc"]

setup_info = {
    "name": "covalent-hpc-plugin",
    "packages": find_packages("."),
    "version": "0.0.6",
    "maintainer": "Andrew S. Rosen",
    "url": "https://github.com/quantum-accelerators/covalent-hpc-plugin",
    "license": "GNU Affero GPL v3.0",
    "author": "Andrew S. Rosen",
    "description": "Covalent HPC Plugin",
    "long_description": open("README.md").read(),
    "long_description_content_type": "text/markdown",
    "include_package_data": True,
    "install_requires": [
        "asyncssh>=2.10.1",
        "aiofiles>=0.8.0",
        "cloudpickle>=2.0.0",
        "covalent>=0.226.0rc0",
        "psij-python>=0.9.0",
    ],
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: Other/Proprietary License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Adaptive Technologies",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development",
        "Topic :: System :: Distributed Computing",
    ],
    "entry_points": {
        "covalent.executor.executor_plugins": plugins_list,
    },
}

if __name__ == "__main__":
    setup(**setup_info)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""HPC executor plugin for the Covalent dispatcher."""

from __future__ import annotations

import asyncio
import datetime
import os
import sys
from pathlib import Path
from typing import Literal, TypedDict

import aiofiles
import asyncssh
import cloudpickle as pickle
import covalent as ct
from aiofiles import os as async_os
from covalent._results_manager.result import Result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent.executor.base import AsyncBaseExecutor

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_EXECUTOR_PLUGIN_DEFAULTS = {
    # SSH credentials
    "address": "",
    "username": "",
    "ssh_key_file": "~/.ssh/id_rsa",
    "cert_file": None,
    # PSI/J parameters
    "instance": "slurm",
    "launcher": "single",
    "inherit_environment": True,
    "environment": {},
    "resource_spec_kwargs": {
        "node_count": 1,
        "processes_per_node": 1,
        "gpu_cores_per_process": 0,
    },
    "job_attributes_kwargs": {
        "duration": 10,  # minutes
        "queue_name": None,
        "project_name": None,
        "custom_attributes": None,
    },
    # Pre/Post-launch commands
    "pre_launch_cmds": [],
    "post_launch_cmds": [],
    "shebang": "#!/bin/bash",
    # Remote Python env parameters
    "remote_python_exe": "python",
    "remote_conda_env": None,
    # Covalent parameters
    "remote_workdir": "~/covalent-workdir",
    "create_unique_workdir": False,
    "cache_dir": str(Path(get_config("dispatcher.cache_dir")).expanduser().resolve()),
    "poll_freq": 60,
    "cleanup": True,
}
_DEFAULT = object()

executor_plugin_name = "HPCExecutor"


class ResourceSpecV1Hint(TypedDict):
    """
    Type hint for PSI/J `ResourceSpecV1` object.

    Reference: https://exaworks.org/psij-python/docs/v/0.9.0/.generated/psij.html#psij.resource_spec.ResourceSpecV1
    """

    node_count: int
    process_count: int
    processes_per_node: int
    cpu_cores_per_process: int
    gpu_cores_per_process: int
    exclusive_node_use: bool


class JobAttributesHint(TypedDict):
    """
    Type hint for PSI/J `JobAttributes` object.

    Reference: https://exaworks.org/psij-python/docs/v/0.9.0/.generated/psij.html#psij.JobAttributes
    """

    duration: datetime.timedelta | int | float # in minutes when type int | float
    queue_name: str | None
    project_name: str | None
    reservation_id: str | None
    custom_attributes: dict[str, object] | None


class HPCExecutor(AsyncBaseExecutor):
    """
    HPC executor plugin class, built around PSI/J.

    This plugin requires that Covalent and PSI/J exist in the remote machine's Python environment.

    Args:
        address: Remote address or hostname of the login node (e.g. "coolmachine.university.edu").
        username: Username used to authenticate over SSH (i.e. what you use to login to `address`).
            The default is None (i.e. no username is required).
        ssh_key_file: Private RSA key used to authenticate over SSH.
            The default is "~/.ssh/id_rsa". If no key is required, set this as None.
        cert_file: Certificate file used to authenticate over SSH, if required (usually has extension .pub).
            The default is None. If no certificate is required, leave this as None.
        instance: The PSI/J `JobExecutor` instance (i.e. job scheduler) to use for job submission.
            Must be one of: "cobalt", "flux", "lsf", "pbspro", "rp", "slurm". Defaults to "slurm".
        launcher: The PSI/J `JobSpec` launcher to use for the job.
            Must be one of: "aprun", "jsrun", "mpirun", "multiple", "single", "srun". Defaults to "single".
        resource_spec_kwargs: The PSI/J keyword arguments for `ResourceSpecV1`, which describes the resources to
            reserve on the scheduling system. Defaults to None, which is equivalent to the PSI/J defaults.
        job_attributes_kwargs: The PSI/J keyword arguments for `JobAttributes`, which describes information about how the
            job is queued and run. Defaults to None, which is equivalent to the PSI/J defaults.
        inherit_environment: Whether the job should inherit the parent environment. Defaults to True.
        environment: Environment variables to set for the job. Defaults to None, which is equivalent to {}.
        pre_launch_cmds: List of shell-compatible commands to run before launching the job. Defaults to None.
        post_launch_cmds: List of shell-compatible commands to run after launching the job. Defaults to None.
        shebang: Shebang to use for pre-launch and post-launch commands. Defaults to "#!/bin/bash".
        remote_python_exe: Python executable to use for job submission. Defaults to "python".
        remote_conda_env: Conda environment to activate on the remote machine. Defaults to None.
        remote_workdir: Working directory on the remote cluster. Defaults to "~/covalent-workdir".
        create_unique_workdir: Whether to create a unique working (sub)directory for each task.
            Defaults to False.
        cache_dir: Local cache directory used by this executor for temporary files.
            Defaults to the dispatcher's cache directory.
        poll_freq: Frequency with which to poll a submitted job. Defaults to 60. Note that settings this value
            to be significantly smaller is not advised, as it will cause too frequent SSHs into the remote machine.
        cleanup: Whether to clean up the temporary job submission files when done. Set this to False for debugging.
            Note that temporary files will be made both in the `remote_workdir` and in `~/.psij`. The latter will
            not be cleaned up by the plugin.
        log_stdout: Path to file to log stdout to. Defaults to "" (i.e. no logging).
        log_stderr: Path to file to log stderr to. Defaults to "" (i.e. no logging).
        time_limit: time limit for the task (in seconds). Defaults to -1 (i.e. no time limit). Note that this is
            not the same as the job scheduler's time limit, which is set in `job_attributes_kwargs`.
        retries: Number of times to retry execution upon failure. Defaults to 0 (i.e. no retries).
    """

    def __init__(
        self,
        # SSH credentials
        address: str = _DEFAULT,
        username: str | None = _DEFAULT,
        ssh_key_file: str | Path | None = _DEFAULT,
        cert_file: str | Path | None = _DEFAULT,
        # PSI/J parameters
        instance: Literal["cobalt", "flux", "lsf", "pbspro", "rp", "slurm"] = _DEFAULT,
        launcher: Literal["aprun", "jsrun", "mpirun", "multiple", "single", "srun"] = _DEFAULT,
        resource_spec_kwargs: ResourceSpecV1Hint = _DEFAULT,
        job_attributes_kwargs: JobAttributesHint = _DEFAULT,
        inherit_environment: bool = _DEFAULT,
        environment: dict[str, str] | None = _DEFAULT,
        # Pre/Post-launch commands
        pre_launch_cmds: list[str] | None = _DEFAULT,
        post_launch_cmds: list[str] | None = _DEFAULT,
        shebang: str | None = _DEFAULT,
        # Remote Python env parameters
        remote_python_exe: str = _DEFAULT,
        remote_conda_env: str | None = _DEFAULT,
        # Covalent parameters
        remote_workdir: str | Path = _DEFAULT,
        create_unique_workdir: bool = _DEFAULT,
        cache_dir: str | Path = _DEFAULT,
        poll_freq: int = _DEFAULT,
        cleanup: bool = _DEFAULT,
        # AsyncBaseExecutor parameters
        log_stdout: str = "",
        log_stderr: str = "",
        time_limit: int = -1,
        retries: int = 0,
    ):
        super().__init__(
            log_stdout=log_stdout, log_stderr=log_stderr, time_limit=time_limit, retries=retries
        )

        config = ct.get_config()
        hpc_config = config["executors"].get("hpc", {})

        # SSH credentials
        self.address = (
            address
            if address != _DEFAULT
            else hpc_config["address"]
            if "address" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["address"]
        )

        self.username = (
            username
            if username != _DEFAULT
            else hpc_config["username"]
            if "username" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["username"]
        )

        self.ssh_key_file = (
            ssh_key_file
            if ssh_key_file != _DEFAULT
            else hpc_config["ssh_key_file"]
            if "ssh_key_file" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["ssh_key_file"]
        )

        self.cert_file = (
            cert_file
            if cert_file != _DEFAULT
            else hpc_config["cert_file"]
            if "cert_file" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["cert_file"]
        )

        # PSI/J parameters
        self.instance = (
            instance
            if instance != _DEFAULT
            else hpc_config["instance"]
            if "instance" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["instance"]
        )

        self.launcher = (
            launcher
            if launcher != _DEFAULT
            else hpc_config["launcher"]
            if "launcher" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["launcher"]
        )

        self.resource_spec_kwargs = (
            resource_spec_kwargs
            if resource_spec_kwargs != _DEFAULT
            else hpc_config["resource_spec_kwargs"]
            if "resource_spec_kwargs" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["resource_spec_kwargs"]
        )

        self.job_attributes_kwargs = (
            job_attributes_kwargs
            if job_attributes_kwargs != _DEFAULT
            else hpc_config["job_attributes_kwargs"]
            if "job_attributes_kwargs" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["job_attributes_kwargs"]
        )

        self.inherit_environment = (
            inherit_environment
            if inherit_environment != _DEFAULT
            else hpc_config["inherit_environment"]
            if "inherit_environment" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["inherit_environment"]
        )

        self.environment = (
            environment
            if environment != _DEFAULT
            else hpc_config["environment"]
            if "environment" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["environment"]
        )

        # Pre/Post-launch commands
        self.pre_launch_cmds = (
            pre_launch_cmds
            if pre_launch_cmds != _DEFAULT
            else hpc_config["pre_launch_cmds"]
            if "pre_launch_cmds" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["pre_launch_cmds"]
        )

        self.post_launch_cmds = (
            post_launch_cmds
            if post_launch_cmds != _DEFAULT
            else hpc_config["post_launch_cmds"]
            if "post_launch_cmds" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["post_launch_cmds"]
        )

        self.shebang = (
            shebang
            if shebang != _DEFAULT
            else hpc_config["shebang"]
            if "shebang" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["shebang"]
        )

        # Remote Python environment parameters
        self.remote_python_exe = (
            remote_python_exe
            if remote_python_exe != _DEFAULT
            else hpc_config["remote_python_exe"]
            if "remote_python_exe" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["remote_python_exe"]
        )

        self.remote_conda_env = (
            remote_conda_env
            if remote_conda_env != _DEFAULT
            else hpc_config["remote_conda_env"]
            if "remote_conda_env" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["remote_conda_env"]
        )

        # Covalent parameters
        self.remote_workdir = (
            remote_workdir
            if remote_workdir != _DEFAULT
            else hpc_config["remote_workdir"]
            if "remote_workdir" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["remote_workdir"]
        )

        self.create_unique_workdir = (
            create_unique_workdir
            if create_unique_workdir != _DEFAULT
            else hpc_config["create_unique_workdir"]
            if "create_unique_workdir" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["create_unique_workdir"]
        )

        self.cache_dir = (
            cache_dir
            if cache_dir != _DEFAULT
            else hpc_config["cache_dir"]
            if "cache_dir" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["cache_dir"]
        )

        self.poll_freq = (
            poll_freq
            if poll_freq != _DEFAULT
            else hpc_config["poll_freq"]
            if "poll_freq" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["poll_freq"]
        )

        self.cleanup = (
            cleanup
            if cleanup != _DEFAULT
            else hpc_config["cleanup"]
            if "cleanup" in hpc_config
            else _EXECUTOR_PLUGIN_DEFAULTS["cleanup"]
        )
        
        if self.poll_freq < 30:
            print("Polling frequency will be increased to 30 seconds.")
            self.poll_freq = 30

        # Make sure local cache dir exists
        os.makedirs(self.cache_dir, exist_ok=True)

    def _format_pickle_script(self) -> str:
        """
        Create the Python script that executes the pickled python function.

        Returns:
            script: String object containing a Python script.
        """

        return f"""
import os
from pathlib import Path

import cloudpickle as pickle

with open(Path(os.path.expandvars("{self._remote_func_filepath}")).expanduser().resolve(), "rb") as f:
    function, args, kwargs = pickle.load(f)

result = None
exception = None

try:
    result = function(*args, **kwargs)
except Exception as e:
    exception = e

with open(Path(os.path.expandvars("{self._remote_result_filepath}")).expanduser().resolve(), "wb") as f:
    pickle.dump((result, exception), f)
"""

    def _format_job_script(self) -> str:
        """
        Create the PSI/J Python script that will submit the compute job to the scheduler.

        NOTE: When PSI/J Remote is released, we can do this directly on the Covalent server
        instead of having to copy the script to the remote machine and running it there.

        Returns:
            String representation of the Python script to make/execute the PSI/J Job object.
        """

        # Fix datetime serialization issue
        if self.job_attributes_kwargs.get("duration") and not isinstance(
            self.job_attributes_kwargs.get("duration"), datetime.timedelta
        ):
            self.job_attributes_kwargs["duration"] = datetime.timedelta(
                minutes=self.job_attributes_kwargs["duration"]
            )

        resources_string = (
            f"resources=ResourceSpecV1(**{self.resource_spec_kwargs}),"
            if self.resource_spec_kwargs
            else ""
        )
        attributes_string = (
            f"attributes=JobAttributes(**{self.job_attributes_kwargs}),"
            if self.job_attributes_kwargs
            else ""
        )
        post_launch_string = (
            f"post_launch=Path(os.path.expandvars('{self._remote_post_launch_filepath}')).expanduser().resolve(),"
            if self.post_launch_cmds
            else ""
        )

        return f"""
import datetime
import os
from pathlib import Path
from psij import Job, JobAttributes, JobExecutor, JobSpec, ResourceSpecV1

job_executor = JobExecutor.get_instance("{self.instance}")

job = Job(
    JobSpec(
        name="{self._name}",
        executable="{self.remote_python_exe}",
        environment={self.environment},
        launcher="{self.launcher}",
        arguments=[str(Path(os.path.expandvars("{self._remote_pickle_script_filepath}")).expanduser().resolve())],
        directory=Path(os.path.expandvars("{self._job_remote_workdir}")).expanduser().resolve(),
        stdout_path=Path(os.path.expandvars("{self._remote_stdout_filepath}")).expanduser().resolve(),
        stderr_path=Path(os.path.expandvars("{self._remote_stderr_filepath}")).expanduser().resolve(),
        pre_launch=Path(os.path.expandvars("{self._remote_pre_launch_filepath}")).expanduser().resolve(),
        {post_launch_string}
        {resources_string}
        {attributes_string}
    )
)

job_executor.submit(job)
native_id = job.native_id
print(native_id)
"""

    def _format_query_status_script(self) -> str:
        """
        Create the PSI/J Python script to query the submitted job status.

        NOTE: When PSI/J Remote is released, we can do this directly on the Covalent server
        instead of having to copy the script to the remote machine and running it there.

        Returns:
            String representation of the Python script to query the job status with PSI/J.
        """

        # NOTE: Once https://github.com/ExaWorks/psij-python/issues/400 is resolved,
        # we can change it to just `target_states=[JobState.QUEUED]`.

        # NOTE: Once https://github.com/ExaWorks/psij-python/issues/401 is resolved,
        # we can remove the `timeout` argument and set `state = job_status.state`.

        return f"""
import datetime
from psij import Job, JobExecutor, JobState

job_executor = JobExecutor.get_instance("{self.instance}")

job = Job()
job_executor.attach(job, f"{self._jobid}")
job_status = job.wait(
    target_states=[
        JobState.QUEUED,
        JobState.CANCELED,
        JobState.ACTIVE,
        JobState.FAILED,
        JobState.COMPLETED,
    ],
    timeout = datetime.timedelta(seconds=10),
)
state = job_status.state or JobState.COMPLETED
print(state.name)
"""

    def _format_pre_launch_script(self) -> str:
        """
        Create the pre-launch script to activate the conda environment,
        check the Python version, and run any user-requested commands.

        Returns:
            String representation of the pre-launch script.
        """
        pre_launch_script = f"{self.shebang}\n" if self.shebang else ""

        if self.remote_conda_env:
            pre_launch_script += f"""
CONDA_BASE=$(conda info --base)
source $CONDA_BASE/etc/profile.d/conda.sh
conda activate {self.remote_conda_env}
retval=$?
if [ $retval -ne 0 ] ; then
    >&2 echo "Conda environment {self.remote_conda_env} is not present on the compute node. "\
    "Please create the environment and try again."
    exit 99
fi
"""
        pre_launch_script += f"""
remote_py_version=$(python -c "print('.'.join(map(str, __import__('sys').version_info[:2])))")
if [[ "{self._remote_python_version}" != $remote_py_version ]] ; then
    >&2 echo "Python version mismatch. Please install Python {self._remote_python_version} in the compute environment."
    exit 199
fi
"""
        if self.pre_launch_cmds:
            pre_launch_script += "".join(f"{cmd}\n" for cmd in self.pre_launch_cmds)

        return pre_launch_script

    def _format_post_launch_script(self) -> str:
        """
        Create the post-launch script to run any user-requested commands.

        Returns:
            String representation of the post-launch script.
        """
        post_launch_script = f"{self.shebang}\n" if self.shebang else ""
        if self.remote_conda_env:
            post_launch_script += f"""
CONDA_BASE=$(conda info --base)
source $CONDA_BASE/etc/profile.d/conda.sh
conda activate {self.remote_conda_env}
"""
        post_launch_script += "".join(f"{cmd}\n" for cmd in self.post_launch_cmds)

        return post_launch_script

    async def _client_connect(self) -> asyncssh.SSHClientConnection:
        """
        Helper function for connecting to the remote host through the asyncssh module.

        Returns:
            The connection object
        """

        if not self.address:
            raise ValueError("address is a required parameter.")

        # Define paths to private key and certificate files
        self.cert_file = Path(self.cert_file).expanduser().resolve() if self.cert_file else None
        self.ssh_key_file = (
            Path(self.ssh_key_file).expanduser().resolve() if self.ssh_key_file else None
        )

        # Read the private key and certificate files
        if self.ssh_key_file:
            if self.cert_file:
                client_keys = [
                    (
                        asyncssh.read_private_key(self.ssh_key_file),
                        asyncssh.read_certificate(self.cert_file),
                    )
                ]
            else:
                client_keys = [asyncssh.read_private_key(self.ssh_key_file)]
        elif self.cert_file:
            raise ValueError("ssh_key_file is required if cert_file is provided.")
        else:
            client_keys = []

        # Connect to the remote host
        try:
            conn = await asyncssh.connect(
                self.address,
                username=self.username,
                client_keys=client_keys,
                known_hosts=None,
            )
        except Exception as e:
            raise RuntimeError(
                f"Could not use asyncssh to connect to host: '{self.address}' as user: '{self.username}'",
                e,
            )

        return conn

    async def run(
        self, function: callable, args: list, kwargs: dict, task_metadata: dict
    ) -> Result:
        """
        Run a function on the remote machine.

        Args:
            function: Function to be executed.
            args: List of positional arguments to be passed to the function.
            kwargs: Dictionary of keyword arguments to be passed to the function.
            task_metadata: Dictionary of metadata associated with the task.

        Returns:
            result: Result object containing the result of the function execution.
        """

        # Set up variables
        dispatch_id = task_metadata["dispatch_id"]
        node_id = task_metadata["node_id"]
        self._name = f"{dispatch_id}-{node_id}"

        results_dir = task_metadata["results_dir"]
        self._task_results_dir = Path(results_dir) / dispatch_id
        self._job_remote_workdir = (
            Path(self.remote_workdir) / dispatch_id / f"node_{node_id}"
            if self.create_unique_workdir
            else Path(self.remote_workdir)
        )

        result_filename = f"result-{dispatch_id}-{node_id}.pkl"
        job_script_filename = f"psij-{dispatch_id}-{node_id}.py"
        pickle_script_filename = f"script-{dispatch_id}-{node_id}.py"
        query_status_script_filename = f"query-status-{dispatch_id}-{node_id}.py"
        func_filename = f"func-{dispatch_id}-{node_id}.pkl"
        stdout_filename = f"stdout-{dispatch_id}-{node_id}.log"
        stderr_filename = f"stderr-{dispatch_id}-{node_id}.log"
        pre_launch_file = f"pre-launch-{dispatch_id}-{node_id}.sh"
        post_launch_file = f"post-launch-{dispatch_id}-{node_id}.sh"

        self._remote_result_filepath = self._job_remote_workdir / result_filename
        self._remote_jobscript_filepath = self._job_remote_workdir / job_script_filename
        self._remote_pickle_script_filepath = self._job_remote_workdir / pickle_script_filename
        self._remote_query_script_filepath = (
            self._job_remote_workdir / query_status_script_filename
        )
        self._remote_func_filepath = self._job_remote_workdir / func_filename
        self._remote_stdout_filepath = self._job_remote_workdir / stdout_filename
        self._remote_stderr_filepath = self._job_remote_workdir / stderr_filename
        self._remote_pre_launch_filepath = self._job_remote_workdir / pre_launch_file
        self._remote_post_launch_filepath = self._job_remote_workdir / post_launch_file

        # Establish connection
        conn = await self._client_connect()

        # Check if the remote Python version is compatible with the local Python version
        self._remote_python_version = ".".join(function.args[0].python_version.split(".")[:2])
        app_log.debug(f"Remote Python version: {self._remote_python_version}")

        # Create the remote directory
        app_log.debug(f"Creating remote work directory: {self._job_remote_workdir}")
        cmd_mkdir_remote = f"mkdir -p {self._job_remote_workdir}"
        proc_mkdir_remote = await conn.run(cmd_mkdir_remote)

        if client_err := proc_mkdir_remote.stderr.strip():
            raise RuntimeError(f"Making remote directory failed: {client_err}")

        # Pickle the function, write to file, and copy to remote filesystem
        async with aiofiles.tempfile.NamedTemporaryFile(dir=self.cache_dir) as temp:
            app_log.debug("Writing pickled function, args, kwargs to file")
            await temp.write(pickle.dumps((function, args, kwargs)))
            await temp.flush()

            app_log.debug(
                f"Copying pickled function to remote filesystem: {self._remote_func_filepath}"
            )
            await asyncssh.scp(temp.name, (conn, self._remote_func_filepath))

        # Format the function execution script, write to file, and copy to remote filesystem
        async with aiofiles.tempfile.NamedTemporaryFile(dir=self.cache_dir, mode="w") as temp:
            python_exec_script = self._format_pickle_script()
            app_log.debug("Writing python run-function script to tempfile")
            await temp.write(python_exec_script)
            await temp.flush()

            app_log.debug(
                f"Copying python run-function to remote filesystem: {self._remote_pickle_script_filepath}"
            )
            await asyncssh.scp(temp.name, (conn, self._remote_pickle_script_filepath))

        # Make the pre launch file
        async with aiofiles.tempfile.NamedTemporaryFile(dir=self.cache_dir, mode="w") as temp:
            pre_launch_script = self._format_pre_launch_script()
            app_log.debug("Writing pre launch file")
            await temp.write(pre_launch_script)
            await temp.flush()

            app_log.debug(
                f"Copying pre launch script to remote filesystem: {self._remote_pre_launch_filepath}"
            )
            await asyncssh.scp(temp.name, (conn, self._remote_pre_launch_filepath))

            cmd_chmod = f"chmod +x {self._remote_pre_launch_filepath}"
            proc_chmod = await conn.run(cmd_chmod)

            if client_err := proc_chmod.stderr.strip():
                raise RuntimeError(f"Changing permissions failed with file: {proc_chmod}")

        if self.post_launch_cmds:
            async with aiofiles.tempfile.NamedTemporaryFile(dir=self.cache_dir, mode="w") as temp:
                post_launch_script = self._format_post_launch_script()
                app_log.debug("Writing post launch file")
                await temp.write(post_launch_script)
                await temp.flush()

                app_log.debug(
                    f"Copying post launch script to remote filesystem: {self._remote_post_launch_filepath}"
                )
                await asyncssh.scp(temp.name, (conn, self._remote_post_launch_filepath))

                cmd_chmod = f"chmod +x {self._remote_post_launch_filepath}"
                proc_chmod = await conn.run(cmd_chmod)

                if client_err := proc_chmod.stderr.strip():
                    raise RuntimeError(f"Changing permissions failed with file: {proc_chmod}")

        # ---------------------------------------------------------------------------------------------
        # NOTE: When PSI/J Remote is released, we can do the following server-side instead.
        # ---------------------------------------------------------------------------------------------

        # Format the PSI/J Python submit script, write to file, and copy to remote filesystem
        async with aiofiles.tempfile.NamedTemporaryFile(dir=self.cache_dir, mode="w") as temp:
            submit_script = self._format_job_script()
            app_log.debug("Writing PSI/J submit script to tempfile")
            await temp.write(submit_script)
            await temp.flush()

            app_log.debug(
                f"Copying PSI/J submit script to remote filesystem: {self._remote_jobscript_filepath} ..."
            )
            await asyncssh.scp(temp.name, (conn, self._remote_jobscript_filepath))

        # Execute the job submission Python script
        app_log.debug("Submitting the job")
        cmd = f"{self.remote_python_exe} {self._remote_jobscript_filepath}"
        if self.remote_conda_env:
            cmd = f"conda activate {self.remote_conda_env} &&" + cmd
        proc = await conn.run(cmd)

        if proc.returncode != 0:
            raise RuntimeError(f"Job submission failed: {proc.stderr.strip()}")

        app_log.debug(f"Job submitted with stdout: {self._remote_stdout_filepath}")
        self._jobid = proc.stdout.strip()

        # Format the Python script to query the job status, write to file, and copy to remote filesystem
        async with aiofiles.tempfile.NamedTemporaryFile(dir=self.cache_dir, mode="w") as temp:
            python_query_status_script = self._format_query_status_script()
            app_log.debug("Writing PSI/J query status script to tempfile")
            await temp.write(python_query_status_script)
            await temp.flush()

            app_log.debug(
                f"Copying PSI/J query status script to remote filesystem: {self._remote_query_script_filepath}"
            )
            await asyncssh.scp(temp.name, (conn, self._remote_query_script_filepath))

        app_log.debug(f"Polling job scheduler with job_id: {self._jobid}")
        await self._poll_scheduler(conn)

        # ---------------------------------------------------------------------------------------------

        app_log.debug(f"Fetching result with job_id: {self._jobid}")
        result, stdout, stderr, exception = await self._fetch_result(conn)

        print(stdout)
        print(stderr, file=sys.stderr)

        if exception:
            raise RuntimeError(f"Fetching job result failed: {stderr}")

        app_log.debug("Preparing for teardown")

        app_log.debug("Closing SSH connection")
        conn.close()
        await conn.wait_closed()
        app_log.debug("SSH connection closed, returning result")

        return result

    async def get_status(self, conn: asyncssh.SSHClientConnection) -> Result | str:
        """
        Query the status of a job previously submitted to the job scheduler.

        Args:
            conn: SSH connection object.

        Returns:
            status: String describing the job status.
        """

        if not hasattr(self, "_jobid"):
            return Result.NEW_OBJ

        cmd = f"{self.remote_python_exe} {self._remote_query_script_filepath}"
        if self.remote_conda_env:
            cmd = f"conda activate {self.remote_conda_env} &&" + cmd
        proc = await conn.run(cmd)

        if proc.returncode != 0:
            raise RuntimeError(f"Getting job status failed: {proc.stderr.strip()}")

        return proc.stdout.strip()

    async def _poll_scheduler(self, conn: asyncssh.SSHClientConnection) -> None:
        """
        Poll for the job status until completion.

        Args:
            conn: SSH connection object.

        Returns:
            None
        """

        # Poll status every `poll_freq` seconds
        status = await self.get_status(conn)

        # Continue to poll until the job is done
        while status in {"QUEUED", "ACTIVE"}:
            await asyncio.sleep(self.poll_freq)
            status = await self.get_status(conn)

        if status != "COMPLETED":
            raise RuntimeError(f"Status for job with native ID {self._jobid}: {status}.")

    async def _fetch_result(
        self,
        conn: asyncssh.SSHClientConnection,
    ) -> tuple[Result, str, str, Exception]:
        """
        Query and retrieve the task result including stdout and stderr logs.

        Args:
            conn: SSH connection object.

        Returns:
            result: Task result.
            stdout: stdout log.
            stderr: stderr log.
            exception: Exception raised during task execution.
        """

        # Check the result file exists on the remote backend
        proc = await conn.run(f"test -e {self._remote_result_filepath}")

        if proc.returncode != 0:
            raise FileNotFoundError(
                proc.returncode, proc.stderr.strip(), self._remote_result_filepath
            )

        # Copy result file from remote machine to Covalent server
        local_result_filename = self._task_results_dir / self._remote_result_filepath.name
        await asyncssh.scp((conn, self._remote_result_filepath), local_result_filename)

        # Copy stdout, stderr from remote machine to Covalent server
        local_stdout_file = self._task_results_dir / self._remote_stdout_filepath.name
        local_stderr_file = self._task_results_dir / self._remote_stderr_filepath.name

        await asyncssh.scp((conn, self._remote_stdout_filepath), local_stdout_file)
        await asyncssh.scp((conn, self._remote_stderr_filepath), local_stderr_file)

        async with aiofiles.open(local_result_filename, "rb") as f:
            contents = await f.read()
            result, exception = pickle.loads(contents)
        await async_os.remove(local_result_filename)

        async with aiofiles.open(local_stdout_file, "r") as f:
            stdout = await f.read()
        await async_os.remove(local_stdout_file)

        async with aiofiles.open(local_stderr_file, "r") as f:
            stderr = await f.read()
        await async_os.remove(local_stderr_file)

        return result, stdout, stderr, exception

    async def teardown(self, task_metadata: dict) -> None:
        """
        Perform cleanup on remote machine.

        Args:
            task_metadata: Dictionary of metadata associated with the task.
                Even though it's not used here, it is always passed by Covalent
                and can't be removed.

        Returns:
            None
        """
        if self.cleanup:
            app_log.debug("Performing cleanup on remote...")
            conn = await self._client_connect()
            await self._perform_cleanup(conn)
    
            app_log.debug("Closing SSH connection...")
            conn.close()
            await conn.wait_closed()
            app_log.debug("SSH connection closed, teardown complete")

    async def _perform_cleanup(self, conn: asyncssh.SSHClientConnection) -> None:
        """
        Function to perform cleanup on remote machine.

        NOTE: When PSI/J Remote is released, the following files will no longer need to be
        deleted: `self._remote_jobscript_filepath`, `self._remote_query_script_filepath`.

        Args:
            conn: SSH connection object

        Returns:
            None
        """
        files_to_remove = [
            self._remote_func_filepath,
            self._remote_pickle_script_filepath,
            self._remote_pre_launch_filepath,
            self._remote_jobscript_filepath,
            self._remote_query_script_filepath,
            self._remote_result_filepath,
            self._remote_stdout_filepath,
            self._remote_stderr_filepath,
        ]
        if self.post_launch_cmds:
            files_to_remove.append(self._remote_post_launch_filepath)
        for f in files_to_remove:
            await conn.run(f"rm {f}")
